/*
#
# ChatBot Eliza - classe de controle
# By Amaury Carvalho
# Em 03/2007
#
*/

#include "eliza.h"

TEliza::TEliza()
{
	srand((unsigned) time(NULL));
	
    sPreviousResponse =  sPreviousInput = "";
    
    text.open("eliza.txt");
    
    endTalk = false;
	
}

TEliza::~TEliza()
{
    text.close();
}

string TEliza::getResponse(string sInput)
{
    string sResponse = "";
    
    cleanString(sInput);
    UpperCase(sInput);
    
    lineIndex = 0;
    
    if(text.opened) {
		if        (sInput == sPreviousInput && sInput.length() > 0) {
            lineIndex = text.searchStart("O");
            if(lineIndex >= 0) sResponse = text.lineText(lineIndex, 2);
		} else if (sInput == "BYE" || sInput == "TCHAU" || sInput == "FIM") {
            lineIndex = text.searchStart("T");
            if(lineIndex >= 0) sResponse = text.lineText(lineIndex, 2);
			endTalk = true;
		} else if (sInput == "WELCOME") {
            string welcome;
            lineIndex = text.searchStart("W");
            while(lineIndex >= 0) {
			   welcome = text.lineText(lineIndex, 2);
			   sResponse += welcome;
               sResponse += "\n";
               lineIndex = text.searchNext();
            } 
		} else {
            int repeat = 0;
            do {
                
               sResponse = find_match(sInput);
           
               if(sResponse == "") {
                   lineIndex = text.searchStart("W");
                   if(lineIndex >= 0) sResponse = text.lineText(lineIndex, 2);
               }
               
               if(sResponse == sPreviousResponse && repeat++ < MAX_RESPONSES) 
                  sResponse = "";
                  
            } while(sResponse == "");
		}
    } else {
       sResponse = "ERRO: BASE DE CONHECIMENTO NAO ENCONTRADA (ELIZA.TXT).";
       endTalk = true;
    }
       
	sPreviousInput = sInput;
	sPreviousResponse = sResponse;

    return sResponse;		
}

string TEliza::find_match(string input)
{
	string keyWord, response = "";
    int lineExact, lineInner, lineGreater, lineLess, lineStart;
    int distGreater, distLess, dist, lineStartLen;
	
    lineIndex = 0;
    lineExact = lineInner = lineGreater = lineLess = lineStart = -1;
    distLess = distGreater = LEV_DIST_MIN;
    
    // procura os tags exatos e proximos (levenshtein)
    
    lineIndex = text.searchStart("P");
    while(lineIndex >= 0) {
        keyWord = text.lineText(lineIndex, 2);
		if(input == keyWord) {
            if(lineExact < 0) lineExact = lineIndex;
		} if(input.find(keyWord) != string::npos ) {
            if(lineInner < 0) lineInner = lineIndex;
		} if((dist = levenshteinDistance(input, keyWord)) < LEV_DIST_MIN) {
                if(input.length() < keyWord.length()) {
                   if(dist < distLess) {
                       lineLess = lineIndex;
                       distLess = dist;
                   }
                } else {
                   if(dist < distGreater < 0) {
                       lineGreater = lineIndex;
                       distGreater = dist;
                   }
                }
        }
		lineIndex = text.searchNext();
	}

    // procura os tags de inicio
    
    lineIndex = text.searchStart("*");
    while(lineIndex >= 0) {
        keyWord = text.lineText(lineIndex, 2);
		if(!strncmp(input.c_str(), keyWord.c_str(), keyWord.length())) {
            lineStart = lineIndex;
            lineStartLen = keyWord.length();
            break;
        }
		lineIndex = text.searchNext();
	}

    // escolhe a melhor resposta
    
    if(lineStart >= 0) 
        response = getResponseFromLine(lineStart) + (char *) (input.c_str()+lineStartLen) + "?";
	else if(lineExact >= 0) 
        response = getResponseFromLine(lineExact);
    else if(lineInner >= 0)
        response = getResponseFromLine(lineInner);
    else if(lineLess >= 0)
        response = getResponseFromLine(lineLess);
    else if(lineGreater >= 0)
        response = getResponseFromLine(lineGreater);
        
    return response;
}

string TEliza::getResponseFromLine(int lineNo)
{
	string responses[MAX_RESPONSES];
    string temp;
    int responsesCount = 0, choosed = 0;
    
    responses[0] = "";
    
    lineIndex = text.searchStart("R", lineNo);
    while(lineIndex >= 0) {
       temp = text.lineText(lineIndex++);
       if(responsesCount < MAX_RESPONSES && temp[0] == 'R') 
          responses[responsesCount++] = (char *) (temp.c_str() + 2);
       else lineIndex = -1;
    }
        
    choosed = rand() % responsesCount;
       
	return responses[choosed];
}
       
//------------------------------------------------------
// Funes de string
//------------------------------------------------------

// torna a string em caixa alta
void TEliza::UpperCase( string &str ) 
{
	int len = str.length();
	int dif = 'a' - 'A';
	for( int i = 0; i < len; i++ ) 
		if ( str[i] >= 'a' && str[i] <= 'z' ) 
			str[i] -= dif;
}

// testa se o caractere  pontuacao
bool TEliza::isPunc(char c) 
{
     const string delim = "?!.;,";
     return delim.find(c) != string::npos;
}

// tira as pontuacoes e espacos redundantes
void TEliza::cleanString( string &str ) 
{
	int len = str.length();
	char prevChar = 0;
	string temp = "";

	for(int i = 0; i < len; ++i) {
		if(str[i] == ' ' && prevChar == ' ') 
			continue;
		else if(!isPunc(str[i])) 
			temp += str[i];

		prevChar = str[i];
	}
	
	str = temp;
}

//------------------------------------------------------
// Clculo da distancia de Levenshtein
//------------------------------------------------------

int TEliza::minimum(int v1, int v2, int v3)
{
    int vm;
    if(v1 <= v2 && v1 <= v3) 
         vm = v1;
    else if(v2 <= v1 && v2 <= v3)
         vm = v2;
    else 
         vm = v3;
    return vm;
}

int TEliza::levenshteinDistance(string &s, string &t)
{
   int d[200][200];
   int i, j, m, n, cost;
 
   m = s.length();
   n = t.length();
 
   for(i=0; i <= m; i++)
       d[i][0] = i;
   for(j=1; j <= n; j++)
       d[0][j] = j;
 
   for(i=1; i <= m; i++)
       for(j=1;j<=n;j++) {
           if(s[i] == t[j]) 
                cost = 0;
           else 
                cost = 1;
                
           d[i][j] = minimum(
                                d[i-1][j] + 1,     // delecao
                                d[i][j-1] + 1,     // insercao
                                d[i-1][j-1] + cost // substituicao
                            );
       }
 
   return d[m][n];
}

